"""Exception types raised in the python API.

.. warning::
   Exception support is still a work in progress. Not all exceptions
   may be raised as :py:exc:`SpglibError`. Please report any exceptions
   that are not caught by :py:exc:`SpglibError`.

.. deprecated:: 2.7.0
   Currently the exception raising is opt-in, requiring to set
   :py:attr:`spglib.error.OLD_ERROR_HANDLING` to ``False``, otherwise
   all exceptions are redirected to :py:func:`spglib.error.get_error_message`.

   For example you could use the following snippet to opt-in this new
   interface.

   .. code-block:: python

       try:
           spglib.error.OLD_ERROR_HANDLING = False
       except AttributeError:
           pass

   or alternatively you can enable it by setting the environment variable
   ``SPGLIB_OLD_ERROR_HANDLING`` to either ``0`` or ``False``.

   Starting from version 2.8.0 the default option will be flipped to always
   throw these exceptions, and the old error handling will be removed in
   version 3.0. Please test out the functionality and provide feedback on the
   GitHub issues.

   Note that :py:attr:`spglib.error.OLD_ERROR_HANDLING` will be removed in a
   future version, so the snippet above is not advised to be used for long-term
   support. Use it only for experimentations and preparing the code base to
   ``try ... except`` the spglib api calls and wait for the 2.8.0 release.

.. versionadded:: 2.7.0
"""

# Copyright (C) 2015 Atsushi Togo
# This file is part of spglib.
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import os
import warnings

from ._compat.warnings import deprecated

__all__ = [
    "SpglibError",
    "get_error_message",
    "OLD_ERROR_HANDLING",
]


class SpglibError(Exception):
    r"""Base exception type for all errors raised by spglib.

    Use this error type in a ``try ... except`` to process any physical or
    other spglib defined errors, such as the number of atom types not matching
    the positions or the distance between atoms being too close. For example:

    .. code-block:: python

       symprec = 1e-10
       while (symprec<1):
           try:
               dataset = spglib.get_symmetry(cell, symprec)
           except spglib.SpglibError as exc:
               # Expected issues by spglib
               # Try again if the symprec was too tight
               if str(exc) == "too close distance between atoms":
                   symprec *= 10
                   continue
               # Otherwise fail gracefully
               print(f"Failed to calculate symmetry:\n{exc}")
               break
           except Exception:
               # Unexpected issues, raise as an unexpected exception
               raise

    .. note::
       More fine-grained exceptions are not defined yet. Please provide
       feedback on what kind of exception separation would be desired.
    """


# TODO: Provide more exception types.


# Old deprecated error functions
OLD_ERROR_HANDLING: bool = True
"""
Use the old error handling.

Note that this variable may be removed in the future or change value in the future.
You can also use :envvar:`SPGLIB_OLD_ERROR_HANDLING` instead of altering this value.
"""


_spglib_error = ""


def _check_OLD_ERROR_HANDLING() -> bool:
    env_var = os.environ.get("SPGLIB_OLD_ERROR_HANDLING")
    if env_var is not None:
        if env_var.lower() in ("false", "0"):
            return False
        return True
    return OLD_ERROR_HANDLING


def _set_or_throw_error(exc: Exception, _throw: bool = False) -> None:
    if _throw or not _check_OLD_ERROR_HANDLING():
        if isinstance(exc, SpglibError):
            # Our native errors we pass transparently
            raise exc
        # Otherwise we try to recast them to SplibError
        raise SpglibError(f"Generic Spglib error:\n{exc}") from exc
    warnings.warn(
        "Set OLD_ERROR_HANDLING to false and catch the errors directly.",
        DeprecationWarning,
        stacklevel=2,
    )
    _spglib_error = str(exc)


def _set_no_error(_throw: bool = False) -> None:
    if _throw or not _check_OLD_ERROR_HANDLING():
        return
    warnings.warn(
        "Set OLD_ERROR_HANDLING to false and catch the errors directly.",
        DeprecationWarning,
        stacklevel=2,
    )
    _spglib_error = "no error"


@deprecated("Set OLD_ERROR_HANDLING to false and catch the errors directly")
def get_error_message() -> str:
    """Return error message why spglib failed.

    .. warning::
        This method is not thread safe, i.e., only safely usable
        when calling one spglib method per process.

    .. versionadded:: 1.9.5
    .. deprecated:: 2.7.0

    """
    return _spglib_error



"""Python bindings for C library for finding and handling crystal."""
# Copyright (C) 2015 Atsushi Togo
# This file is part of spglib.
# SPDX-License-Identifier: BSD-3-Clause

# TODO: _internal should not be exposed, we only continue to do so for
#  backwards compatibility
from ._internal import *  # noqa: F403
from ._internal import __all__ as _internal_all
from ._version import __version__, __version_tuple__
from .cell import *  # noqa: F403
from .cell import __all__ as _cell_all
from .error import *  # noqa: F403
from .error import __all__ as _error_all
from .kpoints import *  # noqa: F403
from .kpoints import __all__ as _kpoints_all
from .msg import *  # noqa: F403
from .msg import __all__ as _msg_all
from .reduce import *  # noqa: F403
from .reduce import __all__ as _reduce_all
from .spg import *  # noqa: F403
from .spg import __all__ as _spg_all
from .utils import *  # noqa: F403
from .utils import __all__ as _utils_all

__all__ = [
    "__version__",
    "__version_tuple__",
    *_internal_all,
    *_cell_all,
    *_error_all,
    *_kpoints_all,
    *_msg_all,
    *_reduce_all,
    *_spg_all,
    *_utils_all,
]



"""Cell operations and standardizations."""

# Copyright (C) 2015 Atsushi Togo
# This file is part of spglib.
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import numpy as np

from . import _spglib
from .error import _set_no_error, _set_or_throw_error
from .spg import SpgCell
from .utils import _expand_cell

__all__ = [
    "standardize_cell",
    "find_primitive",
    "refine_cell",
]


def standardize_cell(
    cell: SpgCell,
    to_primitive: bool = False,
    no_idealize: bool = False,
    symprec: float = 1e-5,
    angle_tolerance: float = -1.0,
) -> SpgCell | None:
    """Return standardized cell. When the search failed, ``None`` is returned.

    Parameters
    ----------
    cell, symprec, angle_tolerance:
        See the docstring of get_symmetry.
    to_primitive : bool
        If True, the standardized primitive cell is created.
    no_idealize : bool
        If True, it is disabled to idealize lengths and angles of basis vectors
        and positions of atoms according to crystal symmetry.

    Returns
    -------
    The standardized unit cell or primitive cell is returned by a tuple of
    (lattice, positions, numbers). If it fails, None is returned.

    Notes
    -----
    .. versionadded:: 1.8

    Now :func:`refine_cell` and :func:`find_primitive` are shorthands of
    this method with combinations of these options.
    About the default choice of the setting, see the documentation of ``hall_number``
    argument of :func:`get_symmetry_dataset`. More detailed explanation is
    shown in the spglib (C-API) document.

    """
    _set_no_error()

    lattice, _positions, _numbers, _ = _expand_cell(cell)

    # Atomic positions have to be specified by scaled positions for spglib.
    num_atom = len(_positions)
    positions = np.zeros((num_atom * 4, 3), dtype="double", order="C")
    positions[:num_atom] = _positions
    numbers = np.zeros(num_atom * 4, dtype="intc")
    numbers[:num_atom] = _numbers
    try:
        num_atom_std = _spglib.standardize_cell(
            lattice,
            positions,
            numbers,
            num_atom,
            int(to_primitive * 1),
            int(no_idealize * 1),
            float(symprec),
            float(angle_tolerance),
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return (
        np.array(lattice.T, dtype="double", order="C"),
        np.array(positions[:num_atom_std], dtype="double", order="C"),
        np.array(numbers[:num_atom_std], dtype="intc"),
    )


def refine_cell(
    cell: SpgCell,
    symprec: float = 1e-5,
    angle_tolerance: float = -1.0,
) -> SpgCell | None:
    """Return refined cell. When the search failed, ``None`` is returned.

    The standardized unit cell is returned by a tuple of
    (lattice, positions, numbers).

    Notes
    -----
    .. versionchanged:: 1.8

    The detailed control of standardization of unit cell can be done using
    :func:`standardize_cell`.

    """
    _set_no_error()

    lattice, _positions, _numbers, _ = _expand_cell(cell)

    # Atomic positions have to be specified by scaled positions for spglib.
    num_atom = len(_positions)
    positions = np.zeros((num_atom * 4, 3), dtype="double", order="C")
    positions[:num_atom] = _positions
    numbers = np.zeros(num_atom * 4, dtype="intc")
    numbers[:num_atom] = _numbers
    try:
        num_atom_std = _spglib.refine_cell(
            lattice,
            positions,
            numbers,
            num_atom,
            float(symprec),
            float(angle_tolerance),
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return (
        np.array(lattice.T, dtype="double", order="C"),
        np.array(positions[:num_atom_std], dtype="double", order="C"),
        np.array(numbers[:num_atom_std], dtype="intc"),
    )


def find_primitive(
    cell: SpgCell,
    symprec: float = 1e-5,
    angle_tolerance: float = -1.0,
) -> SpgCell | None:
    """Primitive cell is searched in the input cell. If it fails, ``None`` is returned.

    The primitive cell is returned by a tuple of (lattice, positions, numbers).

    Notes
    -----
    .. versionchanged:: 1.8

    The detailed control of standardization of unit cell can be done using
    :func:`standardize_cell`.

    """
    _set_no_error()

    lattice, positions, numbers, _ = _expand_cell(cell)

    try:
        num_atom_prim = _spglib.primitive(
            lattice, positions, numbers, float(symprec), float(angle_tolerance)
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return (
        np.array(lattice.T, dtype="double", order="C"),
        np.array(positions[:num_atom_prim], dtype="double", order="C"),
        np.array(numbers[:num_atom_prim], dtype="intc"),
    )



"""These APIs are for the internal usage and development of Spglib itself."""
# Copyright (C) 2015 Atsushi Togo
# This file is part of spglib.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import numpy as np
from numpy._typing import ArrayLike

from . import _spglib
from .error import _set_no_error, _set_or_throw_error
from .spg import SpgCell, SpglibDataset
from .utils import _expand_cell

__all__ = [
    "get_pointgroup",
    "get_symmetry_layerdataset",
    "get_layergroup",
    "get_grid_point_from_address",
    "get_stabilized_reciprocal_mesh",
    "get_grid_points_by_rotations",
    "get_BZ_grid_points_by_rotations",
    "relocate_BZ_grid_address",
]


def get_pointgroup(rotations: ArrayLike[np.intc]) -> tuple[str, int, np.ndarray] | None:
    """Return point group in international table symbol and number.

    The symbols are mapped to the numbers as follows:
    1   "1    "
    2   "-1   "
    3   "2    "
    4   "m    "
    5   "2/m  "
    6   "222  "
    7   "mm2  "
    8   "mmm  "
    9   "4    "
    10  "-4   "
    11  "4/m  "
    12  "422  "
    13  "4mm  "
    14  "-42m "
    15  "4/mmm"
    16  "3    "
    17  "-3   "
    18  "32   "
    19  "3m   "
    20  "-3m  "
    21  "6    "
    22  "-6   "
    23  "6/m  "
    24  "622  "
    25  "6mm  "
    26  "-62m "
    27  "6/mmm"
    28  "23   "
    29  "m-3  "
    30  "432  "
    31  "-43m "
    32  "m-3m "

    """
    _set_no_error()

    # (symbol, pointgroup_number, transformation_matrix)
    try:
        return _spglib.pointgroup(np.array(rotations, dtype="intc", order="C"))
    except Exception as exc:
        _set_or_throw_error(exc)
        return None


def get_symmetry_layerdataset(
    cell: SpgCell,
    aperiodic_dir: int = 2,
    symprec: float = 1e-5,
    _throw: bool = False,
) -> SpglibDataset | None:
    """TODO: Add comments."""
    _set_no_error(_throw)

    lattice, positions, numbers, _ = _expand_cell(cell)

    try:
        spg_ds = _spglib.layer_dataset(
            lattice,
            positions,
            numbers,
            int(aperiodic_dir),
            float(symprec),
        )
    except Exception as exc:
        _set_or_throw_error(exc, _throw)
        return None
    return SpglibDataset(**spg_ds)


def get_layergroup(
    cell: SpgCell,
    aperiodic_dir: int = 2,
    symprec: float = 1e-5,
) -> SpglibDataset | None:
    """Return layer group in ....

    If it fails, None is returned.

    """
    _set_no_error()

    try:
        return get_symmetry_layerdataset(
            cell,
            aperiodic_dir=aperiodic_dir,
            symprec=symprec,
            _throw=True,
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None


def get_grid_point_from_address(
    grid_address: ArrayLike[np.intc],
    mesh: ArrayLike[np.intc],
) -> int | None:
    """Return grid point index by translating grid address."""
    _set_no_error()

    try:
        return _spglib.grid_point_from_address(
            np.array(grid_address, dtype="intc"),
            np.array(mesh, dtype="intc"),
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None


def get_stabilized_reciprocal_mesh(
    mesh: ArrayLike[np.intc],
    rotations: ArrayLike[np.intc],
    is_shift: ArrayLike[np.intc] | None = None,
    is_time_reversal: bool = True,
    qpoints: ArrayLike[np.double] | None = None,
    is_dense: bool = False,
) -> tuple[np.ndarray, np.ndarray] | None:
    """Return k-point map to the irreducible k-points and k-point grid points.

    The symmetry is searched from the input rotation matrices in real space.

    Parameters
    ----------
    mesh : array_like
        Uniform sampling mesh numbers.
        dtype='intc', shape=(3,)
    rotations : array_like
        Rotation matrices with respect to real space basis vectors.
        dtype='intc', shape=(rotations, 3)
    is_shift : array_like
        [0, 0, 0] gives Gamma center mesh and value 1 gives  half mesh shift.
        dtype='intc', shape=(3,)
    is_time_reversal : bool
        Time reversal symmetry is included or not.
    qpoints : array_like
        q-points used as stabilizer(s) given in reciprocal space with respect
        to reciprocal basis vectors.
        dtype='double', shape=(qpoints ,3) or (3,)
    is_dense : bool, optional
        grid_mapping_table is returned with dtype='uintp' if True. Otherwise
        its dtype='intc'. Default is False.

    Returns
    -------
    grid_mapping_table : ndarray
        Grid point mapping table to ir-gird-points.
        dtype='intc', shape=(prod(mesh),)
    grid_address : ndarray
        Address of all grid points. Each address is given by three unsigned
        integers.
        dtype='intc', shape=(prod(mesh), 3)

    """
    _set_no_error()

    if is_dense:
        dtype = "uintp"
    else:
        dtype = "intc"
    mapping_table = np.zeros(np.prod(mesh), dtype=dtype)
    grid_address = np.zeros((np.prod(mesh), 3), dtype="intc")
    if is_shift is None:
        is_shift = [0, 0, 0]
    if qpoints is None:
        qpoints = np.array([[0, 0, 0]], dtype="double", order="C")
    else:
        qpoints = np.array(qpoints, dtype="double", order="C")
        if qpoints.shape == (3,):
            qpoints = np.array([qpoints], dtype="double", order="C")

    try:
        _spglib.stabilized_reciprocal_mesh(
            grid_address,
            mapping_table,
            np.array(mesh, dtype="intc"),
            np.array(is_shift, dtype="intc"),
            int(is_time_reversal * 1),
            np.array(rotations, dtype="intc", order="C"),
            qpoints,
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None
    return mapping_table, grid_address


def get_grid_points_by_rotations(
    address_orig: ArrayLike[np.intc],
    reciprocal_rotations: ArrayLike[np.intc],
    mesh: ArrayLike[np.intc],
    is_shift: ArrayLike[np.intc] | None = None,
    is_dense: bool = False,
) -> np.ndarray | None:
    """Return grid points obtained after rotating input grid address.

    Parameters
    ----------
    address_orig : array_like
        Grid point address to be rotated.
        dtype='intc', shape=(3,)
    reciprocal_rotations : array_like
        Rotation matrices {R} with respect to reciprocal basis vectors.
        Defined by q'=Rq.
        dtype='intc', shape=(rotations, 3, 3)
    mesh : array_like
        dtype='intc', shape=(3,)
    is_shift : array_like, optional
        With (1) or without (0) half grid shifts with respect to grid intervals
        sampled along reciprocal basis vectors. Default is None, which
        gives [0, 0, 0].
    is_dense : bool, optional
        rot_grid_points is returned with dtype='uintp' if True. Otherwise
        its dtype='intc'. Default is False.

    Returns
    -------
    rot_grid_points : ndarray
        Grid points obtained after rotating input grid address
        dtype='intc' or 'uintp', shape=(rotations,)

    """
    _set_no_error()

    if is_shift is None:
        _is_shift = np.zeros(3, dtype="intc")
    else:
        _is_shift = np.array(is_shift, dtype="intc")

    rot_grid_points = np.zeros(len(reciprocal_rotations), dtype="uintp")
    try:
        _spglib.grid_points_by_rotations(
            rot_grid_points,
            np.array(address_orig, dtype="intc"),
            np.array(reciprocal_rotations, dtype="intc", order="C"),
            np.array(mesh, dtype="intc"),
            _is_shift,
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None

    if is_dense:
        return rot_grid_points
    else:
        return np.array(rot_grid_points, dtype="intc")


def get_BZ_grid_points_by_rotations(
    address_orig: ArrayLike[np.intc],
    reciprocal_rotations: ArrayLike[np.intc],
    mesh: ArrayLike[np.intc],
    bz_map: ArrayLike[np.uintp],
    is_shift: ArrayLike[np.intc] | None = None,
    is_dense: bool = False,
) -> np.ndarray | None:
    """Return grid points obtained after rotating input grid address.

    Parameters
    ----------
    address_orig : array_like
        Grid point address to be rotated.
        dtype='intc', shape=(3,)
    reciprocal_rotations : array_like
        Rotation matrices {R} with respect to reciprocal basis vectors.
        Defined by q'=Rq.
        dtype='intc', shape=(rotations, 3, 3)
    mesh : array_like
        dtype='intc', shape=(3,)
    bz_map : array_like
        TODO
    is_shift : array_like, optional
        With (1) or without (0) half grid shifts with respect to grid intervals
        sampled along reciprocal basis vectors. Default is None, which
        gives [0, 0, 0].
    is_dense : bool, optional
        rot_grid_points is returned with dtype='uintp' if True. Otherwise
        its dtype='intc'. Default is False.

    Returns
    -------
    rot_grid_points : ndarray
        Grid points obtained after rotating input grid address
        dtype='intc' or 'uintp', shape=(rotations,)

    """
    _set_no_error()

    if is_shift is None:
        _is_shift = np.zeros(3, dtype="intc")
    else:
        _is_shift = np.array(is_shift, dtype="intc")

    if bz_map.dtype == "uintp" and bz_map.flags.c_contiguous:
        _bz_map = bz_map
    else:
        _bz_map = np.array(bz_map, dtype="uintp")

    rot_grid_points = np.zeros(len(reciprocal_rotations), dtype="uintp")
    try:
        _spglib.BZ_grid_points_by_rotations(
            rot_grid_points,
            np.array(address_orig, dtype="intc"),
            np.array(reciprocal_rotations, dtype="intc", order="C"),
            np.array(mesh, dtype="intc"),
            _is_shift,
            _bz_map,
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None

    if is_dense:
        return rot_grid_points
    else:
        return np.array(rot_grid_points, dtype="intc")


def relocate_BZ_grid_address(
    grid_address: ArrayLike[np.intc],
    mesh: ArrayLike[np.intc],
    reciprocal_lattice: ArrayLike[np.double],  # column vectors
    is_shift: ArrayLike[np.intc] | None = None,
    is_dense: bool = False,
) -> tuple[np.ndarray, np.ndarray] | None:
    """Grid addresses are relocated to be inside first Brillouin zone.

    Number of ir-grid-points inside Brillouin zone is returned.
    It is assumed that the following arrays have the shapes of

    Note that the shape of grid_address is (prod(mesh), 3) and the
    addresses in grid_address are arranged to be in parallelepiped
    made of reciprocal basis vectors. The addresses in bz_grid_address
    are inside the first Brillouin zone or on its surface. Each
    address in grid_address is mapped to one of those in
    bz_grid_address by a reciprocal lattice vector (including zero
    vector) with keeping element order. For those inside first
    Brillouin zone, the mapping is one-to-one. For those on the first
    Brillouin zone surface, more than one addresses in bz_grid_address
    that are equivalent by the reciprocal lattice translations are
    mapped to one address in grid_address. In this case, those grid
    points except for one of them are appended to the tail of this array,
    for which bz_grid_address has the following data storing:

    .. code-block::

      |------------------array size of bz_grid_address-------------------------|
      |--those equivalent to grid_address--|--those on surface except for one--|
      |-----array size of grid_address-----|

    Number of grid points stored in bz_grid_address is returned.
    bz_map is used to recover grid point index expanded to include BZ
    surface from grid address. The grid point indices are mapped to
    (mesh[0] * 2) x (mesh[1] * 2) x (mesh[2] * 2) space (bz_map).

    """
    _set_no_error()

    if is_shift is None:
        _is_shift = np.zeros(3, dtype="intc")
    else:
        _is_shift = np.array(is_shift, dtype="intc")
    bz_grid_address = np.zeros((np.prod(np.add(mesh, 1)), 3), dtype="intc")
    bz_map = np.zeros(np.prod(np.multiply(mesh, 2)), dtype="uintp")
    try:
        num_bz_ir = _spglib.BZ_grid_address(
            bz_grid_address,
            bz_map,
            grid_address,
            np.array(mesh, dtype="intc"),
            np.array(reciprocal_lattice, dtype="double", order="C"),
            _is_shift,
        )
    except Exception as exc:
        _set_or_throw_error(exc)
        return None

    if is_dense:
        return bz_grid_address[:num_bz_ir], bz_map
    else:
        return bz_grid_address[:num_bz_ir], np.array(bz_map, dtype="intc")
